import asyncio
import time

import neovim

from ribosome import command, NvimStatePlugin

from amino.lazy import lazy
from amino import Left, L, _, Map, List, Just, Lists, IO
from amino.state import EvalState
from ribosome.logging import Logging
from ribosome.request import function, msg_function, msg_command
from ribosome.machine import message, may_handle, handle, MachineBase, Message, Nop
from ribosome.machine.state import UnloopedRootMachine, RunScratchMachine, RootMachine
from ribosome.machine.transition import Fatal, may_fallback
from ribosome.machine.scratch import ScratchMachine, Mapping
from ribosome.nvim import NvimFacade, ScratchBuffer
from ribosome.data import Data
from ribosome.record import int_field
from ribosome.machine.base import RunCorosParallel, RunIOsParallel


Msg = message('Msg', 'text')
Err = message('Err')
Scratch = message('Scratch')
ScratchTest = message('ScratchTest')
ScratchCheck = message('ScratchCheck')
St = message('St')
Print = message('Print', 'msg')
RunParallel = message('RunParallel')
RunParallelIOs = message('RunParallelIOs')


class TData(Data):
    v = int_field(initial=0)

    @property
    def _str_extra(self) -> List[str]:
        return List(self.v)


class Mach(Logging):

    @may_handle(Msg)
    def mess(self, data, msg):
        self.log.info(msg.text)
        return data.set(v=3)

    @handle(Err)
    def err(self, data, msg):
        return Left(Fatal(TestPlugin.test_error))

    @may_handle(Scratch)
    def run_scratch(self, data, msg):
        ctor = L(ScratchM)(self.vim, _, _)
        return RunScratchMachine(ctor)

    @may_handle(ScratchCheck)
    def check_scratch(self, data, msg):
        self.log.info(self.sub.length)

    @may_handle(St)
    def st(self, data, msg) -> EvalState[TData, Message]:
        return EvalState.inspect(lambda a: Just(Print(a.v * 2)))

    @may_handle(Print)
    def print_(self, data: Data, msg: Print) -> Message:
        self.log.info(msg.msg)

    @may_handle(RunParallel)
    def run_parallel(self, data: Data, msg: RunParallel) -> Message:
        async def go(n: int) -> None:
            self.log.info(f'sleeping in {n}')
            await asyncio.sleep(.1)
            return Nop().pub
        coros = Lists.range(3) / go
        return RunCorosParallel(coros)

    @may_handle(RunParallelIOs)
    def run_parallel_ios(self, data: Data, msg: RunParallelIOs) -> Message:
        def go(n: int) -> None:
            self.log.info(f'sleeping in {n}')
            time.sleep(.1)
            return Nop().pub
        ios = Lists.range(3) / L(IO.delay)(go, _)
        return RunIOsParallel(ios)


class MachLooped(Mach, RootMachine):
    _data_type = TData


class MachUnlooped(Mach, UnloopedRootMachine):
    _data_type = TData


class ScratchM(ScratchMachine):

    def __init__(self, vim: NvimFacade, scratch: ScratchBuffer,
                 parent: MachineBase) -> None:
        super().__init__(vim, scratch, parent=parent, title='scratch')

    @property
    def prefix(self):
        return 'Test'

    @property
    def mappings(self):
        return Map()

    @may_fallback(ScratchTest)
    def test(self, data, msg):
        self.log.info(TestPlugin.test_scratch)


class TestPlugin(NvimStatePlugin, Logging):
    test_go = 'TestPlugin cmd test message'
    test_fun = 'TestPlugin fun test message'
    test_value = 'test value {}'
    test_error = 'test error'
    test_scratch = 'test scratch'

    def start_plugin(self) -> None:
        self.state().start()

    @neovim.function('Value', sync=True)
    def value(self, args):
        return self.test_value.format(args[0])

    @command(sync=True)
    def go(self):
        self.start_plugin()
        self.log.info(self.test_go)

    @function()
    def fun(self, value):
        return self.test_fun.format(value)

    @msg_function(Msg, sync=True)
    def msg_fun(self):
        pass

    @msg_command(Err)
    def err(self):
        pass

    @msg_command(St)
    def st(self):
        pass

    @msg_command(Scratch)
    def scratch(self):
        pass

    @msg_command(ScratchTest)
    def scratch_test(self):
        pass

    @msg_function(Mapping)
    def test_mapping(self):
        pass

    @msg_command(ScratchCheck)
    def check_scratch(self):
        pass

    @msg_command(RunParallel)
    def run_parallel(self) -> None:
        pass

    @msg_command(RunParallelIOs)
    def run_parallel_i_os(self) -> None:
        pass


class TestPluginLooped(TestPlugin):

    @lazy
    def _state(self) -> MachLooped:
        return MachLooped(self.vim.proxy, title='spec')

    def state(self) -> MachLooped:
        return self._state


class TestPluginUnlooped(TestPlugin):

    @lazy
    def _state(self) -> MachUnlooped:
        return MachUnlooped(self.vim.proxy, title='spec')

    def state(self) -> MachUnlooped:
        return self._state

__all__ = ('TestPlugin',)



import abc
import inspect
from types import FunctionType
from typing import TypeVar, Callable, Union, Any, Dict

from amino import Map, Maybe, Lists, List, Either, Just, Nothing, do
from amino.util.string import camelcaseify, ToStr
from amino.list import Nil

from ribosome.nvim import NvimFacade, NvimIO
from ribosome.logging import ribo_log

A = TypeVar('A')


def rpc_cmd_opt(key: str, value: Union[str, int]) -> Maybe[str]:
    return (
        Just(f'-nargs={value}')
        if key == 'nargs' else
        Nothing
    )


def rpc_cmd_arg(key: str, value: Union[str, int]) -> Maybe[str]:
    return (
        Just(f'[<f-args>]')
        if key == 'nargs' else
        Nothing
    )


def rpc_autocmd_opt(key: str, value: Union[str, int]) -> Maybe[str]:
    return (
        Just(str(value))
        if key == 'pattern' else
        Nothing
    )


RHS = TypeVar('RHS', bound='RpcHandlerSpec')


class RpcHandlerSpec(ToStr):

    @staticmethod
    @do
    def from_spec(spec: Dict[str, Any], method: str, prefix: bool) -> Maybe['RpcHandlerSpec']:
        m = Map(spec)
        ctors = Map(command=RpcCommandSpec, function=RpcFunctionSpec, autocmd=RpcAutocommandSpec)
        tpe = yield m.lift('type')
        Ctor = yield ctors.lift(tpe)
        sync, name, opts = yield m.get_all('sync', 'name', 'opts')
        yield Just(Ctor(sync, name, opts, method, prefix))

    @staticmethod
    def cmd(sync: int, name: str, opts: dict) -> 'RpcHandlerSpec':
        method = f'command:{name}'
        return RpcCommandSpec(sync, name, opts, method, True)

    @staticmethod
    def fun(sync: int, name: str, opts: dict) -> 'RpcHandlerSpec':
        method = f'function:{name}'
        return RpcFunctionSpec(sync, name, opts, method, True)

    @staticmethod
    @do
    def decode(data: dict) -> 'Maybe[RpcHandlerSpec]':
        m = Map(data)
        method, prefix = yield m.get_all('method', 'prefix')
        yield RpcHandlerSpec.from_spec(m, method, prefix)

    def __init__(self, sync: int, name: str, opts: dict, method: str, prefix: bool) -> None:
        self.sync = int(sync) if isinstance(sync, bool) else sync
        self.name = name
        self.opts = opts
        self.method = method
        self.prefix = prefix

    @abc.abstractproperty
    def tpe(self) -> str:
        ...

    @abc.abstractproperty
    def rpc_opts_pre(self) -> List[str]:
        ...

    @abc.abstractproperty
    def rpc_opts_post(self) -> List[str]:
        ...

    @abc.abstractproperty
    def rpc_args(self) -> List[str]:
        ...

    @abc.abstractproperty
    def def_cmd(self) -> str:
        ...

    def _arg_desc(self) -> List[str]:
        return List(self.name, self.tpe, str(self.sync), str(self.opts), self.method, str(self.prefix))

    @property
    def rpc_function(self) -> str:
        return 'rpcrequest' if self.sync else 'rpcnotify'

    def rpc_method(self, plugin_file: str) -> str:
        pre = f'{plugin_file}:' if self.prefix else ''
        return f'{pre}{self.method}'

    @property
    def encode(self) -> dict:
        return dict(
            type=self.tpe,
            sync=self.sync,
            name=self.name,
            opts=self.opts,
            method=self.method,
            prefix=int(self.prefix),
        )


class RpcCommandSpec(RpcHandlerSpec):

    @property
    def tpe(self) -> str:
        return 'command'

    @property
    def rpc_opts_pre(self) -> List[str]:
        return Map(self.opts).map2(rpc_cmd_opt).join

    @property
    def rpc_opts_post(self) -> List[str]:
        return Nil

    @property
    def rpc_args(self) -> List[str]:
        return Map(self.opts).map2(rpc_cmd_arg).join

    @property
    def def_cmd(self) -> str:
        return 'command!'


class RpcFunctionSpec(RpcHandlerSpec):

    @property
    def tpe(self) -> str:
        return 'function'

    @property
    def rpc_opts_pre(self) -> List[str]:
        return Nil

    @property
    def rpc_opts_post(self) -> List[str]:
        return Nil

    @property
    def rpc_args(self) -> List[str]:
        return Nil

    @property
    def def_cmd(self) -> str:
        return 'function!'


class RpcAutocommandSpec(RpcHandlerSpec):

    @property
    def tpe(self) -> str:
        return 'autocmd'

    @property
    def rpc_opts_pre(self) -> List[str]:
        return List()

    @property
    def rpc_opts_post(self) -> List[str]:
        return Map(self.opts).map2(rpc_autocmd_opt).join

    @property
    def rpc_args(self) -> List[str]:
        return Nil

    @property
    def def_cmd(self) -> str:
        return 'autocmd'


def handler(method_name: str, fun: FunctionType) -> Maybe[RpcHandlerSpec]:
    def get(name: str) -> Maybe[A]:
        return Maybe.getattr(fun, name)
    return (
        List('_nvim_rpc_spec', '_nvim_rpc_method_name', '_nvim_prefix_plugin_path')
        .traverse(get, Maybe)
        .flat_map3(RpcHandlerSpec.from_spec)
    )


def register_handler_args(host: str, spec: RpcHandlerSpec, plugin_file: str) -> List[str]:
    fun_prefix = camelcaseify(spec.tpe)
    return List(f'remote#define#{fun_prefix}OnHost', host, spec.rpc_method, spec.sync, spec.name, spec.opts)


def define_handler_native(vim: NvimFacade, host: str, spec: RpcHandlerSpec, plugin_file: str
                          ) -> Either[Exception, None]:
    ribo_log.debug1(lambda: f'defining {spec} on {host}')
    args = register_handler_args(host, spec, plugin_file)
    return vim.call(*args)


class DefinedHandler(ToStr):

    def __init__(self, spec: RpcHandlerSpec, channel: int) -> None:
        self.spec = spec
        self.channel = channel

    def _arg_desc(self) -> List[str]:
        return List(str(self.spec), str(self.channel))


def quote(a: str) -> str:
    return f"'{a}'"


def define_handler_cmd(rhs_f: Callable[[str], str], channel: int, spec: RpcHandlerSpec, plugin_file: str) -> List[str]:
    ribo_log.debug1(lambda: f'defining {spec} on channel {channel}')
    fun = spec.rpc_function
    method = spec.rpc_method(plugin_file)
    args = spec.rpc_args.cons(quote(method)).join_comma
    rpc_call = f'{fun}({channel}, {args})'
    rhs = rhs_f(rpc_call)
    return List(spec.def_cmd) + spec.rpc_opts_pre + List(spec.name) + spec.rpc_opts_post + List(rhs)


def define_handler_io(rhs: Callable[[str], str], channel: int, spec: RpcHandlerSpec, plugin_file: str
                      ) -> NvimIO[DefinedHandler]:
    tokens = define_handler_cmd(rhs, channel, spec, plugin_file)
    return NvimIO.cmd_sync(tokens.join_tokens).replace(DefinedHandler(spec, channel))


def define_function(channel: int, spec: RpcHandlerSpec, plugin_file: str) -> NvimIO[DefinedHandler]:
    return define_handler_io(lambda call: f'(...)\nreturn {call}\nendfunction', channel, spec, plugin_file)


def define_command(channel: int, spec: RpcHandlerSpec, plugin_file: str) -> NvimIO[DefinedHandler]:
    return define_handler_io(lambda a: f'call {a}', channel, spec, plugin_file)


@do
def define_autocmd(channel: int, spec: RpcHandlerSpec, plugin_name: str, plugin_file: str) -> NvimIO[DefinedHandler]:
    yield NvimIO.cmd_sync(f'augroup {plugin_name}')
    result = yield define_handler_io(lambda a: f'call {a}', channel, spec, plugin_file)
    yield NvimIO.cmd_sync(f'augroup end')
    yield NvimIO.pure(result)


def define_handler(channel: int, spec: RpcHandlerSpec, plugin_name: str, plugin_file: str) -> NvimIO[DefinedHandler]:
    if spec.tpe == 'function':
        return define_function(channel, spec, plugin_file)
    elif spec.tpe == 'command':
        return define_command(channel, spec, plugin_file)
    elif spec.tpe == 'autocmd':
        return define_autocmd(channel, spec, plugin_name, plugin_file)
    else:
        return NvimIO.failed(f'invalid type for {spec}')


def rpc_handlers(plugin_class: type) -> List[RpcHandlerSpec]:
    return Lists.wrap(inspect.getmembers(plugin_class)).flat_map2(handler)

__all__ = ('RpcHandlerSpec', 'handler', 'define_handler', 'rpc_handlers')



import neovim

from ribosome.request.base import (RequestHandler, MessageRequestHandler,
                                   JsonMessageRequestHandler)

from amino import Map


class Function(RequestHandler):

    @property
    def desc(self):
        return 'function'

    @property
    def default_kw(self):
        return Map(sync=True)

    @property
    def neovim_fun(self):
        @neovim.function(self.name, **self.kw)
        def neovim_fun_wrapper(obj, *rpc_args):
            return self.dispatch(obj, rpc_args)
        return neovim_fun_wrapper


class MessageFunction(Function, MessageRequestHandler):
    pass


class JsonMessageFunction(Function, JsonMessageRequestHandler):
    pass


def function(**kw):
    def neovim_fun_decorator(fun):
        handler = Function(fun, **kw)
        return handler.neovim_fun
    return neovim_fun_decorator


def msg_function(msg: type, **kw):
    def neovim_msg_fun_decorator(fun):
        handler = MessageFunction(fun, msg, **kw)
        return handler.neovim_fun
    return neovim_msg_fun_decorator


def json_msg_function(msg: type, **kw):
    def neovim_json_msg_fun_decorator(fun):
        handler = JsonMessageFunction(fun, msg, **kw)
        return handler.neovim_fun
    return neovim_json_msg_fun_decorator

__all__ = ('function', 'msg_function', 'json_msg_function')



from typing import TypeVar, Callable, Any, Generic

from amino.tc.base import ImplicitInstances, Implicits
from amino.lazy import lazy
from amino.tc.monad import Monad
from amino import Either, Right, Left, __

from ribosome.nvim.components import NvimComponent

A = TypeVar('A')
B = TypeVar('B')


class NvimIOInstances(ImplicitInstances):

    @lazy
    def _instances(self) -> 'amino.map.Map':
        from amino.map import Map
        return Map({Monad: NvimIOMonad()})


class NvimIO(Generic[A], Implicits, implicits=True, imp_mod='ribosome.nvim.io', imp_cls='NvimIOInstances'):

    @staticmethod
    def wrap_either(f: Callable[[NvimComponent], A]) -> 'NvimIO[A]':
        return NvimIO(lambda a: f(a).get_or_raise)

    @staticmethod
    def cmd_sync(cmdline: str, verbose=False) -> 'NvimIO[str]':
        return NvimIO.wrap_either(__.cmd_sync(cmdline, verbose=verbose))

    @staticmethod
    def call(name: str, *args: Any, **kw: Any) -> 'NvimIO[A]':
        return NvimIO.wrap_either(__.call(name, *args, **kw))

    @staticmethod
    def call_once_defined(name: str, *args: Any, **kw: Any) -> 'NvimIO[A]':
        return NvimIO.wrap_either(__.call_once_defined(name, *args, **kw))

    @staticmethod
    def pure(a: A) -> 'NvimIO[A]':
        return NvimIO(lambda v: a)

    @staticmethod
    def failed(msg: str) -> 'NvimIO[A]':
        def fail() -> A:
            raise Exception(msg)
        return NvimIO(lambda v: fail())

    def __init__(self, run: Callable[[NvimComponent], A]) -> None:
        self.run = run

    def attempt(self, vim: NvimComponent) -> Either[Exception, A]:
        try:
            return Right(self.run(vim))
        except Exception as e:
            return Left(e)

    unsafe_perform_io = attempt

    def effect(self, f: Callable[[A], Any]) -> 'NvimIO[A]':
        def wrap(v: NvimComponent) -> A:
            ret = self.run(v)
            f(ret)
            return ret
        return NvimIO(wrap)

    __mod__ = effect


class NvimIOMonad(Monad[NvimIO]):

    def pure(self, a: A) -> NvimIO[A]:
        return NvimIO.pure(a)

    def flat_map(self, fa: NvimIO[A], f: Callable[[A], NvimIO[B]]) -> NvimIO[B]:
        g = lambda v: f(fa.run(v)).run(v)
        return NvimIO(g)

    def map(self, fa: NvimIO[A], f: Callable[[A], B]) -> NvimIO[B]:
        return NvimIO(lambda a: f(fa.run(a)))

__all__ = ('NvimIO',)



from amino.test.spec_spec import later, IntegrationSpec

from ribosome.test.integration.spec import VimIntegrationSpec, ExternalIntegrationSpec, PluginIntegrationSpec


class VimIntegrationSureHelpers:

    def _log_line(self, index, checker) -> None:
        def check() -> None:
            minlen = index if index >= 0 else abs(index + 1)
            len(self._log_out).should.be.greater_than(minlen)
            return checker(self._log_out[index]).should.be.ok
        later(check)

    def _log_contains(self, line) -> None:
        later(lambda: self._log_out.should.contain(line))

    def _buffer_content(self, data) -> None:
        later(lambda: self.content.should.equal(data))

    def _buffer_length(self, length) -> None:
        later(lambda: self.content.should.have.length_of(length))


class VimIntegrationSpecSpec(VimIntegrationSpec, IntegrationSpec, VimIntegrationSureHelpers):

    def __init__(self) -> None:
        IntegrationSpec.__init__(self)
        VimIntegrationSpec.__init__(self)

    def setup(self) -> None:
        IntegrationSpec.setup(self)
        VimIntegrationSpec.setup(self)

    def teardown(self) -> None:
        IntegrationSpec.teardown(self)
        VimIntegrationSpec.teardown(self)


class ExternalIntegrationSpecSpec(ExternalIntegrationSpec, IntegrationSpec, VimIntegrationSureHelpers):

    def __init__(self) -> None:
        IntegrationSpec.__init__(self)
        ExternalIntegrationSpec.__init__(self)

    def setup(self) -> None:
        IntegrationSpec.setup(self)
        ExternalIntegrationSpec.setup(self)

    def teardown(self) -> None:
        IntegrationSpec.teardown(self)
        VimIntegrationSpec.teardown(self)


class PluginIntegrationSpecSpec(PluginIntegrationSpec, IntegrationSpec, VimIntegrationSureHelpers):

    def __init__(self) -> None:
        IntegrationSpec.__init__(self)
        PluginIntegrationSpec.__init__(self)

    def setup(self) -> None:
        IntegrationSpec.setup(self)
        PluginIntegrationSpec.setup(self)

    def teardown(self) -> None:
        IntegrationSpec.teardown(self)
        VimIntegrationSpec.teardown(self)

__all__ = ('VimIntegrationSureHelpers', 'VimIntegrationSpecSpec', 'ExternalIntegrationSpecSpec',
           'PluginIntegrationSpecSpec')



from amino import Map
from amino.lazy import lazy

from ribosome.data import Data
from ribosome.machine.message_base import Message
from ribosome.machine.interface import MachineI


class Transitions:
    State = Map

    def __init__(self, machine: MachineI, data: Data, msg: Message) -> None:
        self.machine = machine
        self.data = data
        self.msg = msg

    @property
    def name(self):
        return self.machine.title

    @property
    def log(self):
        return self.machine.log

    @lazy
    def local(self):
        if isinstance(self.data, Data):
            return self.data.sub_state(self.name, lambda: self._mk_state)
        else:
            return self._mk_state

    @property
    def _mk_state(self):
        return self.State()

    def with_local(self, new_data):
        return self.data.with_sub_state(self.name, new_data)

__all__ = ('Transitions',)



from typing import Callable, Any
from contextlib import contextmanager

from flexmock import flexmock

from amino import may, Maybe, List
from amino.test.spec_spec import Spec

import ribosome
from ribosome.nvim import Buffer, Tab, Window
from ribosome import NvimFacade
from ribosome.nvim.components import Options, Vars


class MockOptions(Options):

    def set(self, *a, **kw):
        pass


class MockVars(Vars):

    def __init__(self, vim) -> None:
        super().__init__(vim)
        self.vars = dict()
        self.prefix = vim.prefix

    @may
    def __call__(self, name: str) -> Maybe[str]:  # type: ignore
        v = self.vars.get(name)
        if v is None:
            ignore_names = ['_machine', '_message',
                            '{}__message'.format(self.prefix),
                            '{}__machine'.format(self.prefix),
                            ]
            if name not in ignore_names:
                self.log.error('variable not found: {}'.format(name))
        return v

    def set(self, name, value):
        self.vars[name] = value


class MockNvim(object):

    def __init__(self, prefix) -> None:
        self.prefix = prefix
        self._vars = MockVars(self)
        self._options = MockOptions(self)

    @property
    def window(self):
        return MockWindow(self.vim, None, self.prefix)

    @property
    def buffer(self):
        return MockBuffer(self.vim, None, self.prefix)

    @property
    def tab(self):
        return MockTab(self.vim, None, self.prefix)

    def cmd(self, *a, **kw):
        pass


class MockTab(MockNvim, Tab):
    pass


class MockWindow(MockNvim, Window):
    pass


class MockBuffer(MockNvim, Buffer):
    pass


class MockNvimFacade(MockNvim, NvimFacade):

    def __init__(self, prefix):
        MockNvim.__init__(self, prefix)
        self.target = self

    @property
    def windows(self):
        return List()

    def switch_root(self, root):
        pass

    def async(self, f: Callable[['NvimFacade'], Any]):
        return f(self)

    @contextmanager
    def main_event_loop(self):
        yield None

    def reload_windows(self):
        pass


class MockNvimSpec(Spec):

    def __init__(self, prefix):
        Spec.__init__(self)
        self.prefix = prefix

    def setup(self):
        Spec.setup(self)
        ribosome.in_vim = False
        self.vim = MockNvimFacade(self.prefix)
        self.vim_mock = flexmock(self.vim)

__all__ = ('MockNvimSpec', 'MockNvimFacade')



from ribosome.unite.data import (UniteMessage, UniteEntity, UniteSource,
                                 UniteKind)
from ribosome.unite.plugin import mk_unite_candidates, mk_unite_action

__all__ = ('UniteMessage', 'UniteEntity', 'UniteSource', 'UniteKind',
           'mk_unite_candidates', 'mk_unite_action')



from ribosome import NvimStatePlugin

from amino.lazy import lazy
from amino import List, Right, Left, Either, Id, IO
from amino.state import IdState, StateT, EitherState
from ribosome.logging import Logging
from ribosome.request import msg_command
from ribosome.machine import message
from ribosome.machine.state import UnloopedRootMachine, SubMachine, SubTransitions
from ribosome.nvim import NvimFacade, HasNvim, AsyncVimProxy
from ribosome.data import Data
from ribosome.record import field
from ribosome.machine.base import MachineBase
from ribosome.machine import trans

Msg = message('Msg')
Msg2 = message('Msg2')
Msg3 = message('Msg3')
Msg4 = message('Msg4')
Msg5 = message('Msg5')


class Env(Data):
    vim = field(AsyncVimProxy)


class HTrans(SubTransitions, HasNvim, Logging):

    def __init__(self, machine, *a, **kw) -> None:
        SubTransitions.__init__(self, machine, *a, **kw)
        HasNvim.__init__(self, machine.vim)

    @trans.one(Msg, trans.e, trans.st, trans.io)
    def msg(self) -> Either[str, StateT[Id, Env, IO[Msg2]]]:
        return Right(IdState.pure(IO.pure(Msg2())))

    @trans.one(Msg2, trans.e)
    def msg2(self) -> Either[str, Msg]:
        return Left('nothing')

    @trans.unit(Msg3, trans.st)
    def unit(self) -> IdState[Env, Msg]:
        self.log.info('unit')
        return IdState.set(1)

    @trans.one(Msg4, trans.est)
    def est(self) -> EitherState[Env, Msg]:
        return EitherState.pure(Msg5())

    @trans.one(Msg5, trans.est)
    def est_fail(self) -> EitherState[Env, Msg]:
        return EitherState(Left('est'))


class Plugin(SubMachine, HasNvim, Logging):
    Transitions = HTrans

    def __init__(self, vim: NvimFacade, parent=None, title=None) -> None:
        MachineBase.__init__(self, parent, title=title)
        HasNvim.__init__(self, vim)


class Mach(UnloopedRootMachine):
    _data_type = Env

    def __init__(self, vim: NvimFacade, **kw) -> None:
        plug = 'integration._support.handler'
        UnloopedRootMachine.__init__(self, vim, List(plug), **kw)

    @property
    def title(self):
        return 'mach'

    @property
    def init(self):
        return Env(vim=self.vim)


class HandlerSpecPlugin(NvimStatePlugin, Logging, name='handler'):

    @lazy
    def _state(self) -> Mach:
        return Mach(self.vim.proxy, title='handler')

    def state(self) -> Mach:
        return self._state

    def start_plugin(self) -> None:
        self.state().start()

    @msg_command(Msg, sync=True)
    def msg(self):
        pass

    @msg_command(Msg3, sync=True)
    def unit(self):
        pass

    @msg_command(Msg4, sync=True)
    def est(self):
        pass

__all__ = ('HandlerSpecPlugin',)



from amino import Map, L, _, List, Left, Right, __

from ribosome.util.callback import VimCallback, parse_callback_spec


class TransitionHelpers:

    @property
    def options(self):
        return Map()

    def _callback_errors(self, spec, cb):
        def log(s, e):
            self.log.error('failed to parse callback \'{}\': {}'.format(s, e))
        (spec.zip(cb)).map2(lambda s, c: c.lmap(L(log)(s, _)))

    def _inst_callback(self, name, target):
        t = target or self.vim
        return (name(t)
                if isinstance(name, type) and issubclass(name, VimCallback)
                else name)

    def _inst_callbacks(self, spec, target, special):
        cb = (
            spec /
            L(parse_callback_spec)(_, special) /
            __.flat_map(__.func(self.vim))
        )
        self._callback_errors(spec, cb)
        return cb.filter(_.is_right).join / L(self._inst_callback)(_, target)

    def _callback(self, name, target=None, special=Map()):
        spec = self.options.get(name).o(self.vim.vars.p(name))
        return self._inst_callbacks(spec, target, special)

    def _callbacks(self, name, target=None, special=Map()):
        var = self.vim.vars.pl(name) | List()
        opt = self.options.get(name) | List()
        return self._inst_callbacks(var + opt, target, special)

    def _from_opt(self, tpe, **strict):
        o = self.options ** Map(strict)
        missing = tpe.mandatory_fields.k.filter_not(o.has_key)
        msg = 'cannot create {} without params: {}'
        return (Left(msg.format(tpe, missing)) if missing else
                Right(tpe.from_opt(o)))

__all__ = ('TransitionHelpers',)



from amino.options import EnvOption

development = EnvOption('RIBOSOME_DEVELOPMENT')
spec = EnvOption('RIBOSOME_SPEC')
file_log_level = EnvOption('RIBOSOME_FILE_LOG_LEVEL')
file_log_fmt = EnvOption('RIBOSOME_FILE_LOG_FMT')
nvim_log_file = EnvOption('NVIM_PYTHON_LOG_FILE')

__all__ = ('development', 'spec', 'file_log_level', 'file_log_fmt', 'niivm_log_file')
